/**
 ** fingersExecTaskCodels.c
 **
 ** Codels called by execution task fingersExecTask
 **
 ** Author: 
 ** Date: 
 **
 **/

#ifdef VXWORKS
# include <vxWorks.h>
#else
# include <portLib.h>
#endif
#include "server/fingersHeader.h"

#include "ftdi.h"
#include "fingersConst.h"

#include <jidogrip.h>
#include <commonStructLib.h>
#include <errno.h>

typedef enum BOOLEAN {false, true} BOOLEAN;
static struct ftdi_context ftdi;
struct GRIP_STR *grip = NULL;
static FINGERS_CMD_STR oldCmd;	/* store the current state of commands */
static int moduleIsInitialised = 0;
static int nbcle = 0;
static  double position;

static STATUS fingersConnectUSB(void);
static STATUS fingersDisconnectUSB(void);
static STATUS fingersAcquisitionStart(void);
static STATUS fingersAcquisitionMain(void);
static STATUS fingersAcquisitionEnd(void);

ACTIVITY_EVENT fingersDisconnectUSBMain(int *report);

/*------------------------------------------------------------------------
 *
 * fingersExecTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
fingersExecTaskInit(int *report)
{
  moduleIsInitialised = 0;
  return OK;
}

/*------------------------------------------------------------------------
 *
 * fingersExecTaskEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
fingersExecTaskEnd(void)
{
  int status = 0;

  if (moduleIsInitialised == 0) {
    /* not initialised correctly - nothing to do*/
    return OK;
  }
  if (fingersAcquisitionEnd() != 0)
    status++;
  if (fingersDisconnectUSB() != 0)
    status++;
  moduleIsInitialised = 0;
  if (status == 0)
    return OK;
  else
    return ERROR;
}

/*------------------------------------------------------------------------
 *
 * fingersExecTaskPerm  --  codel of permanent activity
 *
 * Description: 
 * 
 * Reports:      OK
 *              S_fingers_CANNOT_CONNECT_USB
 *              S_fingers_CANNOT_START_READ
 *              S_fingers_READ_PB
 * 
 * Returns:    OK or ERROR
 */

STATUS
fingersExecTaskPerm(int *report)
{
  unsigned char status; 
  FINGERS_CMD_STR *command;

  if (moduleIsInitialised == 0) {
    /* not initialized yet - nothing to do */
    return OK;
  }

  command = &SDI_F->command;

  /* Test if commands changed, send appropriate command if needed */
  if (oldCmd.command != command->command) {
    if (command->command == FINGERS_OPEN) {
      jidogripOpen(grip);
    } else {
      jidogripClose(grip);
    }
  } 
  if (oldCmd.speed != command->speed) {
    if (command->speed == FINGERS_FAST) {
      jidogripFast(grip);
    } else {
      jidogripSlow(grip);
    }
  }
  if (oldCmd.stop != command->stop) {
    if (command->stop == FINGERS_TRUE) {
      jidogripStop(grip);
    } else {
      jidogripStart(grip);
    }
  }
  /* store current state */
  memcpy(&oldCmd, command, sizeof(FINGERS_CMD_STR));

  /* update status */
  status = jidogripStatus(grip);
  SDI_F->status.open = !!(status & GRIP_OPEN_INPUT);
  SDI_F->status.closed = !!(status & GRIP_CLOSE_INPUT);
  SDI_F->status.active = !(status & GRIP_STOP_INPUT);

//  if (fingersAcquisitionMain() != OK) {
//    *report = S_fingers_READ_PB;
//    fprintf(stderr, "%s:%d: Pb: fingersAcquisitionMain()\n",
//	    __FILE__, __LINE__);
//    /*    return ERROR; */
//    return OK;
//  }


  SDI_F->cpt =  SDI_F->cpt +1;
  if ( SDI_F->cpt > TIMEOUT_TO_GRASP) {
    SDI_F->cpt =0;
  }
  
  if (SDI_F->mode == FINGERS_GRASPING_MODE) {
    if ( SDI_F->status.closed == TRUE ) {
   //  if ( (SDI_F->status.closed == TRUE) && (SDI_F->ExchangeObjectTerminated == 1)) {
      SDI_F->command.command = FINGERS_OPEN;
      SDI_F->grasped = 0;
    }
  }


//  /*  printf("mode %d\n",SDI_F->mode); */
//  if (SDI_F->mode == FINGERS_GRASPING_MODE) {
//    if ( SDI_F->measures.position < MIN_THICKNESS_OF_OBJECT && (SDI_F->ExchangeObjectTerminated == 1)) {
//      SDI_F->command.command = FINGERS_OPEN;
//      SDI_F->grasped = 0;
//    }
//  }




  return OK;
}

/* ============================== STATIC ============================== */
static STATUS fingersConnectUSB(void)
{
//  int status = 0;

//  status = ftdi_init(&ftdi);
//  if (status < 0) {
//    fprintf(stderr, "%s:%d: Pb: Couldn't allocate read buffer\n",
//	    __FILE__, __LINE__);
//    return -1;
//  } 
//  status = ftdi_usb_open(&ftdi, 0x0403, 0x6001); 
//  if(status != 0) {
//    fprintf(stderr,
//	    "%s:%d: Pb: unable to open ftdi device\nstatus= %d errno= %d %s\n",
//	    __FILE__, __LINE__, status, errno, strerror(errno));
//    return -1;
//  }
  printf("fingers : FTDI open succeeded\n");
  if (grip == NULL) {
    grip = jidogripInit(FINGERS_DIO_DEV_NAME);
    if (grip == NULL) {
      return -1;
    }
  }
  return OK;
}

static STATUS fingersDisconnectUSB(void)
{
  int status = 0;

  status = ftdi_usb_close(&ftdi);
  if (status < 0) {
    fprintf(stderr, "%s:%d: Pb: couldn't close FTDI device\n",
	    __FILE__, __LINE__);
    return -1;
  }
  ftdi_deinit(&ftdi);
  fprintf(stderr, "ftdi close succeeded\n");
  jidogripEnd(grip);
  grip = NULL;
  return OK;
}

static STATUS fingersAcquisitionStart(void)
{
  /* start */
  char buf[] = "START";
  int total_written = 0;

  total_written = ftdi_write_data(&ftdi,buf,sizeof(buf));
  /* start not ok */
  if (total_written != sizeof(buf)) {
    fprintf(stderr, "%s:%d: Pb: couldn't close FTDI device\n",
	    __FILE__, __LINE__);
    return ERROR;
  }
  printf ("fingersAcquisitionStart: START sent\n");
  return OK;
}


static STATUS fingersAcquisitionMain(void)
{
  static unsigned char buf[FINGERS_MAXBUFSIZE];
  unsigned char tab[2] = {0x00,0x00};
  unsigned char Packet_ID = 0;
  unsigned char *pRebuild = tab;
  unsigned char divider = 0;
  unsigned int reconstruction = 0;

  int size_buf;
  int i = 0;

  //struct Sensors Sensors_Result, Sensors_cancellation;
  BOOLEAN DLE_bit = 0;
  BOOLEAN dividande = 0;
  //BOOLEAN cancellation = 0;

  commonStructGive ((void *) fingersDataStrId);
  commonStructGive ((void *) fingersCntrlStrId);    

  size_buf = ftdi_read_data(&ftdi, buf, FINGERS_MAXBUFSIZE);

  if (commonStructTake (fingersCntrlStrId) != OK ||
      commonStructTake (fingersDataStrId) != OK) {
    logMsg("fingersExecTask: commonStructTake error\n");
    return ERROR;
  }

  if (size_buf < 0) {
    perror("fingersAcquisitionMain");
    fprintf(stderr, "%s:%d: Pb ftdi_read_data: size_buf = %d\n",
	    __FILE__, __LINE__, size_buf);
    return ERROR;
  }
  /*   fprintf (stdout, "s=%d  ", size_buf); */
  /*   fflush(stdout); */
  for (i = 0; i < size_buf; i++) {
    if (buf[i] == 0x10) {
      if(DLE_bit == 0) {
	DLE_bit = 1;
      } else {
	DLE_bit = 0;
	*pRebuild++ = buf[i];
      }
    } else {
      if(DLE_bit == 1) {
	// header identifier
	// dividande reception (2 bytes -> int)
	if (buf[i] >= 0x11 && buf[i] <= 0x19) {
	  Packet_ID = buf[i];		// identifiant du capteur
	  pRebuild = tab;	        // zone d'enregistrement du dividande
	  dividande = 1;		// dividande reception in progress
	} else {
	  // divider reception (1 byte -> unsigned char)
	  //pRebuild = tab + 4;		// pRebuild = tab + 4 => tab[4]
	  //divider = 1;
	  pRebuild = &divider;		// pRebuild = tab + 4 => tab[4]
	}
      } else {
	//c'est un mot normal
	if (Packet_ID != 0)
	  *pRebuild++ = buf[i];
      }
      DLE_bit = 0;
    } /* if DLE_bit != 0 */
    if (dividande == 1 && divider != 0) {
      reconstruction = tab[0] +(256 * tab[1]);
      switch (Packet_ID) {
      case 0x11	:		
	SDI_F->measures.strainGauge[0] =
	  0.0583 * (double)reconstruction / (double)divider - 1.2086;
	if(SDI_F->measures.strainGauge[0] <= 0.0)
	  SDI_F->measures.strainGauge[0] = 0.0;
	break;
      case 0x12	:	
	SDI_F->measures.strainGauge[1] =
	  0.0523 * (double)reconstruction / (double)divider - 0.9801;
	if(SDI_F->measures.strainGauge[1] <= 0.0)
	  SDI_F->measures.strainGauge[1] = 0.0;
	break;
      case 0x13	:
	SDI_F->measures.fsr[0] =
	  0.4250 * (double)reconstruction / (double)divider - 166.4186;
	if(SDI_F->measures.fsr[0] <= 0.0)
	  SDI_F->measures.fsr[0] = 0.0;
	break;
      case 0x14	:
	SDI_F->measures.fsr[1] =
	  0.1778 * (double)reconstruction / (double)divider - 68.4140;
	if(SDI_F->measures.fsr[1] <= 0.0)
	  SDI_F->measures.fsr[1] = 0.0;
	break;
      case 0x15	:
	SDI_F->measures.fsr[2] =
	  0.2118 * (double)reconstruction / (double)divider - 85.2449;
	if(SDI_F->measures.fsr[2] <= 0.0)
	  SDI_F->measures.fsr[2] = 0.0;
	break;
      case 0x16	:
	SDI_F->measures.fsr[3] =
	  0.2489 * (double)reconstruction / (double)divider - 95.5715;
	if(SDI_F->measures.fsr[3] <= 0.0)
	  SDI_F->measures.fsr[3] = 0.0;
	break;
      case 0x17	:
	SDI_F->measures.fsr[4] =
	  0.2147 * (double)reconstruction / (double)divider - 87.2645;
	if(SDI_F->measures.fsr[4] <= 0.0)
	  SDI_F->measures.fsr[4] = 0.0;
	break;
      case 0x18	:
	SDI_F->measures.fsr[5] =
	  0.2147 * (double)reconstruction / (double)divider - 87.2645;
	if(SDI_F->measures.fsr[5] <= 0.0)
	  SDI_F->measures.fsr[5] = 0.0;
	break;
      case 0x19	:
	SDI_F->measures.position =
	  -0.000085837 * (double)reconstruction / (double)divider + 0.0635;
	if(SDI_F->measures.position <= 0.0)
	  SDI_F->measures.position = 0.0;
	break;
      default     :   
	break;
      } /* dividande == 1 && divider != 0 */
      
      dividande = 0;
      divider = 0;
      Packet_ID = 0;
      reconstruction = 0;
    }
  } /* for */
  
  return OK;
}

static STATUS fingersAcquisitionEnd(void)
{
  char buf[] = "STOP";
  int total_written = 0;

  total_written = ftdi_write_data(&ftdi,buf,sizeof(buf));
   /* stop not ok */
  if (total_written != sizeof(buf)) 
    return -1;
  return OK;
}
/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_fingers_CANNOT_CONNECT_USB
 *              S_fingers_CANNOT_START_READ
 *              S_fingers_READ_PB
 */

/* fingersInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT 
fingersInitMain(FINGERS_INIT_PARAM *(in_initParam), int *report)

{
  SDI_F->grasped = in_initParam->grasped;
  SDI_F->mode =  in_initParam->mode;
  SDI_F->ExchangeObjectTerminated = 1;
  /* current state unkown */
  if (SDI_F->grasped == 1) {
    oldCmd.command = -1;
    SDI_F->command.command = FINGERS_CLOSE;
  } else {
    oldCmd.command = -1;
  }
  oldCmd.speed = -1;
  oldCmd.stop = -1;
  
  if (fingersConnectUSB() != OK) {
    *report = S_fingers_CANNOT_CONNECT_USB;
    return ETHER;
  }
//  if (fingersAcquisitionStart() != OK) {
//    *report = S_fingers_CANNOT_START_READ;
//    return ETHER;
//  }
  SDI_F->cpt = 0;
  moduleIsInitialised = 1;
  return ETHER;
}

/*------------------------------------------------------------------------
 * Stop
 *
 * Description: 
 *
 * Reports:      OK
 *              S_fingers_CloseError
 */

/* fingersStopMain  -  codel EXEC of Stop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
fingersStopMain(int *report)
{
  if (fingersExecTaskEnd() == ERROR) {
    *report = S_fingers_CloseError;
  }
  return ETHER;
}



/*------------------------------------------------------------------------
 * ExchangeObjectHumanVsRobot
 *
 * Description: 
 *
 * Reports:      OK
 *              S_fingers_fingersExchangeObject
 */

/* fingersExchangeObjectHumanVsRobotStart  -  codel START of ExchangeObjectHumanVsRobot
   Returns:  START EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT 
fingersExchangeObjectHumanVsRobotStart(int *(out_grasped), int *report)
{
  if (SDI_F->grasped == 0) {   
    SDI_F->command.command = FINGERS_OPEN;     
  }
  SDI_F->cpt =0;
  nbcle =0; 
  SDI_F->ExchangeObjectTerminated = 0;
  *(out_grasped) = SDI_F->grasped;
  return EXEC;
}

/* fingersExchangeObjectHumanVsRobotMain  -  codel EXEC of ExchangeObjectHumanVsRobot
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT 
fingersExchangeObjectHumanVsRobotMain(int *(out_grasped), int *report)

{
  if ((nbcle == 0) && (SDI_F->cpt > 10)) {
    if (((SDI_F->measures.fsr[5]) > THRESHOLD_TO_GRASP) && (SDI_F->grasped == 0)) {   
      SDI_F->command.command = FINGERS_CLOSE;
      nbcle = 1;
      SDI_F->cpt =0;
      position = SDI_F->measures.position;
    }
    if (((SDI_F->measures.fsr[5]) > THRESHOLD_TO_GRASP) && (SDI_F->grasped == 1)) {
      SDI_F->command.command = FINGERS_OPEN;
      SDI_F->grasped = 0;
      *(out_grasped) = 0;
      nbcle = 0;
      SDI_F->cpt =0;
      SDI_F->ExchangeObjectTerminated = 1;
      return ETHER;
    }
    if  ((SDI_F->measures.position < MIN_THICKNESS_OF_OBJECT) && ( SDI_F->grasped == 1)) {
      SDI_F->command.command = FINGERS_OPEN;
      SDI_F->grasped = 0;
	  *(out_grasped) = 0;
      nbcle = 0;
      SDI_F->cpt =0;
      SDI_F->ExchangeObjectTerminated = 0;
    }
  }
  if ((nbcle == 1) &&  (SDI_F->cpt > 10)) {
      if ( (position - SDI_F->measures.position) < 0.001) {
	if ( (SDI_F->measures.position) < MIN_THICKNESS_OF_OBJECT) {
	  SDI_F->command.command = FINGERS_OPEN;
	  SDI_F->grasped = 0;
	  *(out_grasped) = 0;
	  nbcle = 0;
	  SDI_F->cpt =0;
	  SDI_F->ExchangeObjectTerminated = 0;
	}
	else {
	  SDI_F->grasped = 1;
	  nbcle = 0;
	  SDI_F->cpt =0;
	  *(out_grasped) = 1;
	  SDI_F->ExchangeObjectTerminated = 1;
	  return ETHER;
	}
      }
      else {
	position = SDI_F->measures.position;
	SDI_F->ExchangeObjectTerminated = 0;
      }  
  }
/*   printf("Boucle %d\n", nbcle); */
/*   printf("Grasped %d\n", SDI_F->grasped); */
/*   printf("Grasped out %d\n", *(out_grasped) ); */
/*   printf("Position %f\n", SDI_F->measures.position); */
  SDI_F->ExchangeObjectTerminated = 0;
  return EXEC;
}





/*------------------------------------------------------------------------
 * AutoExchangeObjectHumanVsRobot
 *
 * Description: 
 *
 * Reports:      OK
 *              S_fingers_fingersAutoExchangeObject
 */

/* fingersAutoExchangeObjectHumanVsRobotStart  -  codel START of AutoExchangeObjectHumanVsRobot
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
fingersAutoExchangeObjectHumanVsRobotStart(int *report)
{
   if (SDI_F->grasped == 0) {   
    SDI_F->command.command = FINGERS_OPEN;     
  }
  SDI_F->cpt =0;
  nbcle =0;  
  return EXEC;
}

/* fingersAutoExchangeObjectHumanVsRobotMain  -  codel EXEC of AutoExchangeObjectHumanVsRobot
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
fingersAutoExchangeObjectHumanVsRobotMain(int *report)
{
   if ((nbcle == 0) && (SDI_F->cpt > 10)) {
    if (((SDI_F->measures.fsr[5]) > THRESHOLD_TO_GRASP) && (SDI_F->grasped == 0)) {   
      SDI_F->command.command = FINGERS_CLOSE;
      nbcle = 1;
      SDI_F->cpt =0;
      position = SDI_F->measures.position;
    }
    if (((SDI_F->measures.fsr[5]) > THRESHOLD_TO_GRASP) && (SDI_F->grasped == 1)) {
      SDI_F->command.command = FINGERS_OPEN;
      SDI_F->grasped = 0; 
      nbcle = 0;
      SDI_F->cpt =0;
    }
    if  ((SDI_F->measures.position < MIN_THICKNESS_OF_OBJECT) && ( SDI_F->grasped == 1)) {
      SDI_F->command.command = FINGERS_OPEN;
      SDI_F->grasped = 0;
      nbcle = 0;
      SDI_F->cpt =0;
    }
   }
   if ((nbcle == 1) &&  (SDI_F->cpt > 10)) {
     if ( (position - SDI_F->measures.position) < 0.001) {
       if ( (SDI_F->measures.position) < MIN_THICKNESS_OF_OBJECT) {
	 SDI_F->command.command = FINGERS_OPEN;
	 SDI_F->grasped = 0;
	 nbcle = 0;
	 SDI_F->cpt =0;
       }
       else {
	 SDI_F->grasped = 1;
	 nbcle = 0;
	 SDI_F->cpt =0;
       }
     }
     else {
       position = SDI_F->measures.position;
     }  
   }
   /*   printf("Boucle %d\n", nbcle); */
   /*   printf("Grasped %d\n", SDI_F->grasped); */
   /*   printf("Position %f\n", SDI_F->measures.position); */
   return EXEC;
}




